# window.py
#
# Copyright 2024 Etersoft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import sys
import gi
import subprocess
import threading

from gi.repository import Gtk, Adw, GLib


class CommandRunner:
    def __init__(self, on_done=None):
        self.dialog = None
        self.textbuffer = None
        
        self.on_done = on_done  # Callback function to be called after completion

    def run_command(self, command, dialog):
        self.dialog = dialog
        self.textbuffer = dialog.get_child().get_child().get_child().get_buffer()

        def append_log(text):
            GLib.idle_add(self.textbuffer.insert_at_cursor, text)
            GLib.idle_add(dialog.get_child().get_child().get_child().scroll_to_mark,
                          self.textbuffer.get_insert(), 0, False, 0.0, 1.0)

        def read_output(source):
            while True:
                line = source.readline()
                if line:
                    GLib.idle_add(append_log, line)
                else:
                    break

        def process_runner():
            try:
                process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE, text=True)
                threading.Thread(target=read_output, args=(process.stdout,)).start()
                threading.Thread(target=read_output, args=(process.stderr,)).start()
                
                process.wait()
                
                GLib.idle_add(dialog.close)
                
                if self.on_done:
                    GLib.idle_add(self.on_done)  # Call the callback function after completion
            
            except subprocess.CalledProcessError as e:
                GLib.idle_add(append_log, f"Ошибка: {str(e)}\n")
                GLib.idle_add(dialog.close)
                if self.on_done:
                    GLib.idle_add(self.on_done)  # Call the callback function after completion

        threading.Thread(target=process_runner).start()

class ApplicationManager:
    @staticmethod
    def get_available_applications(callback):
        threading.Thread(target=ApplicationManager._get_available_applications, args=(callback,)).start()

    @staticmethod
    def _get_available_applications(callback):
        process = subprocess.Popen(["epm", "play", "--list-all"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        def read_output():
            stdout, stderr = process.communicate()
            if process.returncode == 0:
                applications = ApplicationManager.parse_applications_output(stdout)
                GLib.idle_add(callback, applications)
            else:
                GLib.idle_add(callback, [], stderr)  # Pass error message

        threading.Thread(target=read_output).start()

    @staticmethod
    def get_installed_applications(callback):
        threading.Thread(target=ApplicationManager._get_installed_applications, args=(callback,)).start()

    @staticmethod
    def _get_installed_applications(callback):
        process = subprocess.Popen(["epm", "play", "--list"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        def read_output():
            stdout, stderr = process.communicate()
            if process.returncode == 0:
                installed_apps = ApplicationManager.parse_installed_applications_output(stdout)
                GLib.idle_add(callback, installed_apps)
            else:
                GLib.idle_add(callback, [], stderr)  # Pass error message

        threading.Thread(target=read_output).start()

    @staticmethod
    def parse_applications_output(output):
        applications = []
        
        for line in output.splitlines()[1:]:  # Пропустить заголовок
            parts = line.split(' - ', 1)  # Ограничить сплит до 2 частей
            if len(parts) == 2:
                name, dscr = map(str.strip, parts)
                applications.append({
                    'name': name.replace('&', '&amp;'),
                    'dscr': dscr.replace('&', '&amp;')
                })
            else:
                # Логгирование некорректной строки для диагностики
                print(f"Неправильный формат строки: {line}")
        
        return applications

    @staticmethod
    def parse_installed_applications_output(output):
        lines = output.splitlines()[1:]
        return [line.split(' - ')[0].strip().split()[0] for line in lines]


@Gtk.Template(resource_path='/ru/eepm/PlayGUI/window.ui')
class EepmPlayGuiWindow(Adw.ApplicationWindow):
    __gtype_name__ = 'EepmPlayGuiWindow'
    search_entry = Gtk.Template.Child()
    search_dropdown = Gtk.Template.Child()

    loading_spinner = Gtk.Template.Child()
    choice_listbox = Gtk.Template.Child()
    apply_button = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.apply_button.connect("activated", self.on_apply_clicked)
        
        self.search_entry.connect("search-changed", self.on_search_changed)
        self.search_dropdown.connect("notify::selected", self.on_filter_changed)
        
        self.choice_listbox.set_filter_func(self.listbox_filter_func)
        
        self.loading_count = 0  # Track loading operations
        self.load_applications()

    def show_loading_spinner(self):
        self.loading_spinner.set_visible(True)
        self.choice_listbox.set_visible(False)  # Скрыть группу настроек во время загрузки
        self.loading_count += 1  # Increment loading counter

    def hide_loading_spinner(self):
        self.loading_count -= 1  # Decrement loading counter
        if self.loading_count == 0:
            self.loading_spinner.set_visible(False)
            self.choice_listbox.set_visible(True)

    def load_applications(self):
        self.show_loading_spinner()
        ApplicationManager.get_available_applications(self.on_applications_loaded)

    def on_applications_loaded(self, applications, error=None):
        if error:
            print(f"Error: {error}")
        else:
            self.applications = applications
            self.update_ui()
        self.hide_loading_spinner()  # Move this to ensure it waits for installation loading too

    def update_ui(self):
        self.show_loading_spinner()  # Show loading again for installed apps
        ApplicationManager.get_installed_applications(self.on_installed_apps_loaded)
        self.choice_listbox.invalidate_filter()

    def on_installed_apps_loaded(self, installed_apps, error=None):
        if error:
            print(f"Error: {error}")
        else:
            self.installed_apps = installed_apps
            self.clear_choice_listbox()

            self.checkboxes = {}
            for app in self.applications:
                self.add_application_row(app)

        self.hide_loading_spinner()  # Now only hides after both loads are done

    def clear_choice_listbox(self):
        self.choice_listbox.remove_all()

    def add_application_row(self, app):
        row = Adw.ActionRow(
            title=app['name'],
            subtitle=app['dscr']
        )
        
        checkbox = Gtk.CheckButton()
        checkbox.add_css_class("selection-mode")
        
        checkbox.set_active(app['name'] in self.installed_apps)
        
        checkbox.connect("toggled", self.on_checkbox_toggled, app['name'])
        
        row.add_suffix(checkbox)
        self.choice_listbox.append(row)
        self.checkboxes[app['name']] = checkbox

    def on_checkbox_toggled(self, checkbox, app_name):
        active = checkbox.get_active()
        print(f"{app_name} {'установлен' if active else 'снят'}")
    
    def on_search_changed(self, search_entry):
        self.choice_listbox.invalidate_filter()  # Обновление фильтра при изменении поиска

    def on_filter_changed(self, dropdown, _pspec):
        self.choice_listbox.invalidate_filter()  # Обновление фильтра при изменении фильтра

    def listbox_filter_func(self, row):
        """Функция фильтрации для GtkListBox, которая проверяет текст и состояние фильтра"""
        search_text = self.search_entry.get_text().lower()
        filter_option = self.search_dropdown.get_selected()

        # Получение заголовка и подзаголовка строки
        title = row.get_title().lower() if row.get_title() else ''
        subtitle = row.get_subtitle().lower() if row.get_subtitle() else ''

        # Проверка текста поиска (по имени или описанию)
        matches_search = search_text in title or search_text in subtitle

        # Проверка по фильтру: установлено/не установлено в системе
        app_name = row.get_title() if row.get_title() else ''
        is_installed = app_name in self.installed_apps

        if filter_option == 1:  # Installed
            return matches_search and is_installed
        elif filter_option == 2:  # Uninstalled
            return matches_search and not is_installed
        else:
            return matches_search  # All

    def on_apply_clicked(self, button):
        self.show_loading_spinner()  # Показать сообщение о загрузке перед выполнением команды
        to_install, to_remove = self.get_install_remove_lists()
        commands = self.build_commands(to_install, to_remove)

        if commands:
            full_command = " && ".join(commands)
            pkexec_command = f'pkexec sh -c "{full_command}"'
            self.show_log_dialog(pkexec_command)

    def get_install_remove_lists(self):
        to_install = [app_name for app_name, checkbox in self.checkboxes.items() if checkbox.get_active() and app_name not in self.installed_apps]
        to_remove = [app_name for app_name, checkbox in self.checkboxes.items() if not checkbox.get_active() and app_name in self.installed_apps]
        return to_install, to_remove

    def build_commands(self, to_install, to_remove):
        commands = []
        if to_install:
            commands.append(f"epm play --auto {' '.join(to_install)}")
        if to_remove:
            commands.append(f"epm play --auto --remove {' '.join(to_remove)}")
        return commands

    def show_log_dialog(self, command):
        dialog = Adw.Dialog(title="Лог выполнения")
        dialog.set_follows_content_size(True)
        
        dialog_clamp = Adw.Clamp(
            margin_bottom=12,
            margin_end=12,
            margin_top=12,
            margin_start=12,
            maximum_size=500
        )
        
        textview = Gtk.TextView()
        textview.set_editable(False)
        textbuffer = textview.get_buffer()

        scrolled_window = Gtk.ScrolledWindow(
            width_request=304,
            height_request=259
        )
        scrolled_window.set_child(textview)
        scrolled_window.set_vexpand(True)
        
        dialog_clamp.set_child(scrolled_window)

        dialog.set_child(dialog_clamp)
        dialog.present(self)

        # Создание и передача функции обратного вызова для обновления UI
        runner = CommandRunner(on_done=self.update_ui_after_log)
        runner.run_command(command, dialog)

    def update_ui_after_log(self):
        self.update_ui()  # Обновление UI после закрытия диалога
        self.hide_loading_spinner()  # Скрыть сообщение о загрузке после закрытия диалога

